/*
 * sdl.c
 * sdl interfaces -- based on svga.c 
 *
 * (C) 2001 Damian Gryski <dgryski@uwaterloo.ca>
 * Joystick code contributed by David Lau
 * Sound code added by Laguna
 *
 * Licensed under the GPLv2, or later.
 */

#include <stdlib.h>
#include <stdio.h>

#include <SDL.h>


#include "fb.h"
#include "input.h"
#include "rc.h"

struct fb fb;

static int use_yuv = -1;
static int fullscreen = 1;
static int use_altenter = 1;
static int use_joy = 1, sdl_joy_num;
static SDL_Joystick * sdl_joy = NULL;
static const int joy_commit_range = 3276;
static char Xstatus, Ystatus;
static int grab_input;

/* skin */
static char *skindir;
static char *skin_background;
static SDL_Surface *bgimage;

static SDL_Surface *screen;
static SDL_Overlay *overlay;
static SDL_Rect overlay_rect;

static int vmode[3] = { 0, 0, 16 };

rcvar_t vid_exports[] =
{
	RCV_VECTOR("vmode", &vmode, 3),
	RCV_BOOL("yuv", &use_yuv),
	RCV_BOOL("fullscreen", &fullscreen),
	RCV_BOOL("altenter", &use_altenter),
    RCV_BOOL("grabinput", &grab_input),

    RCV_STRING("skindir", &skindir),
    RCV_STRING("skin_background", &skin_background),
	RCV_END
};

rcvar_t joy_exports[] =
{
	RCV_BOOL("joy", &use_joy),
	RCV_END
};

int skin_load(const char *dirname)
{
    int size;
    char filename[FILENAME_MAX];
    char *skinrc = NULL;

    snprintf(filename, sizeof(filename), "%s/skin.rc", dirname);
    fprintf(stderr, "skinrc filename = %s\n", filename);
    rc_sourcefile(filename);

    snprintf(filename, sizeof(filename), "%s/%s", dirname, skin_background);
    fprintf(stderr, "skin_background filename = %s\n", filename);
    bgimage = SDL_LoadBMP(filename);
    if (bgimage == NULL) {
        fprintf(stderr, "skin_load: %s\n", SDL_GetError());
        return 0;
    }
    fprintf(stderr, "skin_load: successed.\n");

    return 1;
}

/* keymap - mappings of the form { scancode, localcode } - from sdl/keymap.c */
extern int keymap[][2];

static int mapscancode(SDLKey sym)
{
	/* this could be faster:  */
	/*  build keymap as int keymap[256], then ``return keymap[sym]'' */

	int i;
	for (i = 0; keymap[i][0]; i++)
		if (keymap[i][0] == sym)
			return keymap[i][1];
	if (sym >= '0' && sym <= '9')
		return sym;
	if (sym >= 'a' && sym <= 'z')
		return sym;
	return 0;
}


static void joy_init()
{
	int i;
	int joy_count;
	
	/* Initilize the Joystick, and disable all later joystick code if an error occured */
	if (!use_joy) return;
	
	if (SDL_InitSubSystem(SDL_INIT_JOYSTICK))
		return;
	
	joy_count = SDL_NumJoysticks();
	
	if (!joy_count)
		return;

	/* now try and open one. If, for some reason it fails, move on to the next one */
	for (i = 0; i < joy_count; i++)
	{
		sdl_joy = SDL_JoystickOpen(i);
		if (sdl_joy)
		{
			sdl_joy_num = i;
			break;
		}	
	}
	
	/* make sure that Joystick event polling is a go */
	SDL_JoystickEventState(SDL_ENABLE);
}

static void overlay_init()
{
	if (!use_yuv) return;
	
	if (use_yuv < 0)
		if (vmode[0] < 320 || vmode[1] < 288)
			return;
	
	overlay = SDL_CreateYUVOverlay(320, 144, SDL_YUY2_OVERLAY, screen);

	if (!overlay) return;

	if (!overlay->hw_overlay || overlay->planes > 1)
	{
		SDL_FreeYUVOverlay(overlay);
		overlay = 0;
		return;
	}

	SDL_LockYUVOverlay(overlay);
	
	fb.w = 160;
	fb.h = 144;
	fb.pelsize = 4;
	fb.pitch = overlay->pitches[0];
	fb.ptr = overlay->pixels[0];
	fb.yuv = 1;
	fb.cc[0].r = fb.cc[1].r = fb.cc[2].r = fb.cc[3].r = 0;
	fb.dirty = 1;
	fb.enabled = 1;
	
	overlay_rect.x = 0;
	overlay_rect.y = 0;
	overlay_rect.w = vmode[0];
	overlay_rect.h = vmode[1];

	/* Color channels are 0=Y, 1=U, 2=V, 3=Y1 */
	switch (overlay->format)
	{
		/* FIXME - support more formats */
	case SDL_YUY2_OVERLAY:
	default:
		fb.cc[0].l = 0;
		fb.cc[1].l = 24;
		fb.cc[2].l = 8;
		fb.cc[3].l = 16;
		break;
	}
	
	SDL_UnlockYUVOverlay(overlay);
}

void vid_init()
{
	int flags;

	if (!vmode[0] || !vmode[1])
	{
		int scale = rc_getint("scale");
		if (scale < 1) scale = 1;
		vmode[0] = 178/*160*/ * scale;
		vmode[1] = 144 * scale;
	}

	flags = SDL_ANYFORMAT | SDL_HWPALETTE | SDL_HWSURFACE;

/* 	if (fullscreen) */
/* 		flags |= SDL_FULLSCREEN; */

	if (SDL_Init(SDL_INIT_VIDEO))
		die("SDL: Couldn't initialize SDL: %s\n", SDL_GetError());

	if (!(screen = SDL_SetVideoMode(vmode[0], vmode[1], vmode[2], flags)))
		die("SDL: can't set video mode: %s\n", SDL_GetError());

	if (vmode[0] == 320 || vmode[1] == 240) {
        skin_load(skindir);
	}

	SDL_ShowCursor(0);
    if (grab_input)
        SDL_WM_GrabInput(SDL_GRAB_ON);

	joy_init();

	overlay_init();

	if (fb.yuv) return;

	SDL_LockSurface(screen);

/* 	fb.w = screen->w; */
/* 	fb.h = screen->h; */
	fb.w = 160;
	fb.h = 144;
	fb.pelsize = 2;						/* only supported 16bpp */
	fb.pitch = screen->pitch / 2;
	fb.indexed = 0;
/* 	fb.indexed = fb.pelsize == 0; */
/* 	fb.ptr = screen->pixels; */
	fb.ptr = (Uint16 *)screen->pixels +
		((screen->h - 144) / 2 * (screen->pitch / 2)) +
		(screen->w - 178) / 2 + 1;
	fb.cc[0].r = screen->format->Rloss;
	fb.cc[0].l = screen->format->Rshift;
	fb.cc[1].r = screen->format->Gloss;
	fb.cc[1].l = screen->format->Gshift;
	fb.cc[2].r = screen->format->Bloss;
	fb.cc[2].l = screen->format->Bshift;

	SDL_UnlockSurface(screen);

	fb.enabled = 1;
	fb.dirty = 1;

/* 	printf("%d %d %d %d\n", fb.w, fb.h, fb.pelsize, fb.pitch); */
}


void ev_poll()
{
	event_t ev;
	SDL_Event event;
	int axisval;

    static int prev_pen_axis[2] = { 0, 0 };

	while (SDL_PollEvent(&event))
	{
		switch(event.type)
		{
		case SDL_ACTIVEEVENT:
			if (event.active.state == SDL_APPACTIVE)
				fb.enabled = event.active.gain;
			break;
		case SDL_KEYDOWN:
			if ((event.key.keysym.sym == SDLK_RETURN) && (event.key.keysym.mod & KMOD_ALT))
				SDL_WM_ToggleFullScreen(screen);
			ev.type = EV_PRESS;
			ev.code = mapscancode(event.key.keysym.sym);
			ev_postevent(&ev);
			break;
		case SDL_KEYUP:
			ev.type = EV_RELEASE;
			ev.code = mapscancode(event.key.keysym.sym);
			ev_postevent(&ev);
			break;
		case SDL_JOYAXISMOTION:
			switch (event.jaxis.axis)
			{
			case 0: /* X axis */
				axisval = event.jaxis.value;
				if (axisval > joy_commit_range)
				{
					if (Xstatus==2) break;
					
					if (Xstatus==0)
					{
						ev.type = EV_RELEASE;
						ev.code = K_JOYLEFT;
        			  		ev_postevent(&ev);				 		
					}
					
					ev.type = EV_PRESS;
					ev.code = K_JOYRIGHT;
					ev_postevent(&ev);
					Xstatus=2;
					break;
				}	   				   
				
				if (axisval < -(joy_commit_range))
				{
					if (Xstatus==0) break;
					
					if (Xstatus==2)
					{
						ev.type = EV_RELEASE;
						ev.code = K_JOYRIGHT;
        			  		ev_postevent(&ev);				 		
					}
					
					ev.type = EV_PRESS;
					ev.code = K_JOYLEFT;
					ev_postevent(&ev);
					Xstatus=0;
					break;
				}	   				   
				
				/* if control reaches here, the axis is centered,
				 * so just send a release signal if necisary */
				
				if (Xstatus==2)
				{
					ev.type = EV_RELEASE;
					ev.code = K_JOYRIGHT;
					ev_postevent(&ev);
				}
				
				if (Xstatus==0)
				{
					ev.type = EV_RELEASE;
					ev.code = K_JOYLEFT;
					ev_postevent(&ev);
				}	       
				Xstatus=1;
				break;
				
			case 1: /* Y axis*/ 
				axisval = event.jaxis.value;
				if (axisval > joy_commit_range)
				{
					if (Ystatus==2) break;
					
					if (Ystatus==0)
					{
						ev.type = EV_RELEASE;
						ev.code = K_JOYUP;
        			  		ev_postevent(&ev);				 		
					}
					
					ev.type = EV_PRESS;
					ev.code = K_JOYDOWN;
					ev_postevent(&ev);
					Ystatus=2;
					break;
				}	   				   
				
				if (axisval < -joy_commit_range)
				{
					if (Ystatus==0) break;
					
					if (Ystatus==2)
					{
						ev.type = EV_RELEASE;
						ev.code = K_JOYDOWN;
        			  		ev_postevent(&ev);
					}
					
					ev.type = EV_PRESS;
					ev.code = K_JOYUP;
					ev_postevent(&ev);
					Ystatus=0;
					break;
				}	   				   
				
				/* if control reaches here, the axis is centered,
				 * so just send a release signal if necisary */
				
				if (Ystatus==2)
				{
					ev.type = EV_RELEASE;
					ev.code = K_JOYDOWN;
					ev_postevent(&ev);
				}
				
				if (Ystatus==0)
				{
					ev.type = EV_RELEASE;
					ev.code = K_JOYUP;
					ev_postevent(&ev);
				}
				Ystatus=1;
				break;
			}
			break;
		case SDL_JOYBUTTONUP:
			if (event.jbutton.button>15) break;
			ev.type = EV_RELEASE;
			ev.code = K_JOY0 + event.jbutton.button;
			ev_postevent(&ev);
			break;
		case SDL_JOYBUTTONDOWN:
			if (event.jbutton.button>15) break;
			ev.type = EV_PRESS;
			ev.code = K_JOY0+event.jbutton.button;
			ev_postevent(&ev);
			break;
        case SDL_MOUSEBUTTONDOWN:
            if ((screen->w >> 3) * 5 < event.button.x)
            {
                ev.type = EV_PRESS;
                prev_pen_axis[0] = ev.code = K_PENRIGHT;
                ev_postevent(&ev);
            }
            else if ((screen->w >> 3) * 3 > event.button.x)
            {
                ev.type = EV_PRESS;
                prev_pen_axis[0] = ev.code = K_PENLEFT;
                ev_postevent(&ev);
            }
            if ((screen->h >> 3) * 5 < event.button.y)
            {
                ev.type = EV_PRESS;
                prev_pen_axis[1] = ev.code = K_PENDOWN;
                ev_postevent(&ev);
            }
            else if ((screen->h >> 3) * 3 > event.button.y)
            {
                ev.type = EV_PRESS;
                prev_pen_axis[1] = ev.code = K_PENUP;
                ev_postevent(&ev);
            }

            if (!(prev_pen_axis[0] || prev_pen_axis[1]))
            {
                if ((screen->w >> 1) > event.button.x)
                {
                    ev.type = EV_PRESS;
                    prev_pen_axis[0] = ev.code = K_PENCENTERLEFT;
                    ev_postevent(&ev);
                }
                else
                {
                    ev.type = EV_PRESS;
                    prev_pen_axis[0] = ev.code = K_PENCENTERRIGHT;
                    ev_postevent(&ev);
                }
            }

            break;
        case SDL_MOUSEBUTTONUP:
            if (prev_pen_axis[0]) {
                ev.type = EV_RELEASE;
                ev.code = prev_pen_axis[0];
                prev_pen_axis[0] = 0;
                ev_postevent(&ev);
            }
            if (prev_pen_axis[1]) {
                ev.type = EV_RELEASE;
                ev.code = prev_pen_axis[1];
                prev_pen_axis[1] = 0;
                ev_postevent(&ev);
            }

            break;
		case SDL_QUIT:
			exit(1);
			break;
		default:
			break;
		}
	}
}

void vid_setpal(int i, int r, int g, int b)
{
	SDL_Color col;

	col.r = r; col.g = g; col.b = b;

	SDL_SetColors(screen, &col, i, 1);
}

void vid_preinit()
{
}

void vid_close()
{
	if (overlay)
	{
		SDL_UnlockYUVOverlay(overlay);
		SDL_FreeYUVOverlay(overlay);
	}
	else SDL_UnlockSurface(screen);
	SDL_Quit();
	fb.enabled = 0;
}

void vid_settitle(char *title)
{
	SDL_WM_SetCaption(title, title);
}

void vid_begin()
{
	static int bgRenderd = 0;
	if (bgimage != NULL && bgRenderd < 2) {
		bgRenderd ++;
		SDL_BlitSurface(bgimage, 0, screen, 0);
		SDL_UpdateRect(screen, 0, 0, screen->w, screen->h);
	}

	if (overlay)
	{
		SDL_LockYUVOverlay(overlay);
		fb.ptr = overlay->pixels[0];
		return;
	}
	SDL_LockSurface(screen);
/* 	fb.ptr = screen->pixels; */
}

#define FRAME_INTERVAL 16

void vid_end()
{
	if (overlay)
	{
		SDL_UnlockYUVOverlay(overlay);
		if (fb.enabled)
			SDL_DisplayYUVOverlay(overlay, &overlay_rect);
		return;
	}
	SDL_UnlockSurface(screen);

	{
		static Uint32 nowTick;
		static Uint32 prvTickCount = 0;
		int frame;
		nowTick = SDL_GetTicks();
		frame = (int)(nowTick - prvTickCount) / FRAME_INTERVAL;
		if (frame <= 0) {
			frame = 1;
			SDL_Delay(prvTickCount + FRAME_INTERVAL - nowTick);
			prvTickCount += FRAME_INTERVAL;
		} else if (frame > 5) {
			frame = 5;
			prvTickCount = nowTick;
		} else {
			prvTickCount += frame * FRAME_INTERVAL;
		}

/* 		if (fb.enabled) SDL_Flip(screen); */
		if (fb.enabled) {
#if 1
			SDL_UpdateRect(screen,
						   (screen->w - 160) >> 1, (screen->h - 144) >> 1,
						   160, 144);
#else
			SDL_Rect rl = {0, 0, 8, 144};
			SDL_Rect rr = {168, 0, 8, 144};
			SDL_FillRect(screen, &rl, 0);
			SDL_FillRect(screen, &rr, 0);
			SDL_UpdateRect(screen, 0, 0, 0, 0);
#endif
		}
	}
}





#include "pcm.h"


#ifdef USE_SDL_AUDIO

struct pcm pcm;


static int sound = 1;
static int samplerate = 44100;
static int stereo = 0;
static volatile int audio_done;

rcvar_t pcm_exports[] =
{
	RCV_BOOL("sound", &sound),
	RCV_INT("stereo", &stereo),
	RCV_INT("samplerate", &samplerate),
	RCV_END
};


#define RING_SIZE 4						/* must be 2^x */
#define RING_MASK (RING_SIZE - 1)

static void *audioBuffer[RING_SIZE];
static int readPtr, writePtr;
static int inlen = 0;

static void audio_callback(void *blah, byte *stream, int len)
{
/* 	fprintf(stderr, "audio_callback: len=0x%x\n", len); */
/* 	fprintf(stderr, "readPtr=%d, writePtr=%d\n", readPtr, writePtr); */
	if (inlen == 0) {
/* 		fprintf(stderr, "audio buffer is empty\n"); */
		return;
	}
	SDL_MixAudio(stream, audioBuffer[readPtr], len, SDL_MIX_MAXVOLUME);
/* 	memcpy(stream, audioBuffer[readPtr], len); */
	readPtr = (readPtr + 1) & RING_MASK;
	--inlen;
}


#define NSAMPLE 0x800

/* static unsigned char buf[SOUND_BUF_LEN] __attribute__ ((aligned (32))); */

void pcm_init()
{
	int i;
	SDL_AudioSpec as;

	if (!sound) return;

	SDL_InitSubSystem(SDL_INIT_AUDIO);
	as.freq = samplerate;
	as.format = AUDIO_S16;
	as.channels = 1;//1 + stereo;
/* 	as.samples = samplerate / 60; */
/* 	for (i = 1; i < as.samples; i<<=1); */
/* 	as.samples = i; */
	as.samples = NSAMPLE;
	as.callback = audio_callback;
	as.userdata = 0;
	if (SDL_OpenAudio(&as, 0) == -1)
		return;

	for (i = 0; i < RING_SIZE; ++i) {
		audioBuffer[i] = malloc(as.size);
		memset(audioBuffer[i], 0, as.size);
	}
	readPtr = 0, writePtr = 0;

	pcm.hz = as.freq;
	pcm.stereo = as.channels - 1;
	pcm.len = NSAMPLE;//0x2000;//as.size;
	pcm.buf = audioBuffer[0];
	pcm.pos = 0;
/* 	memset(pcm.buf, 0, pcm.len); */
/* 	fprintf(stderr, "as.freq = %d, pcm.len = 0x%x pcm.stereo=%d, as.size=0x%x\n", */
/* 			as.freq, pcm.len, pcm.stereo, as.size); */

	SDL_PauseAudio(1);
}



int pcm_submit()
{
	if (!pcm.buf) return 0;
	if (pcm.pos < pcm.len) return 1;
	while (!audio_done)
		SDL_Delay(4);
	audio_done = 0;
	pcm.pos = 0;
	return 1;
}

void pcm_close()
{
	if (sound) {
		int i;

		SDL_CloseAudio();
		for (i = 0; i < RING_SIZE; ++i)
			free(audioBuffer[i]);
	}
}

void render_pcm(int len)
{
/* 	fprintf(stderr, "render_pcm: pcm.len = 0x%x, pcm.pos = 0x%x\n", */
/* 			pcm.len, pcm.pos); */
	if (SDL_GetAudioStatus() != SDL_AUDIO_PLAYING) {
		if (inlen > RING_SIZE / 2)
			SDL_PauseAudio(0);
	}

	if (inlen == RING_SIZE) {
/* 		fprintf(stderr, "audio buffer is full\n"); */
		return;
	}

	pcm.buf = audioBuffer[writePtr];

	SDL_LockAudio();
	writePtr = (writePtr + 1) & RING_MASK;
	++inlen;
	SDL_UnlockAudio();

	sound_mix(pcm.len - pcm.pos);
/* 	pcm_submit(); */
	pcm.pos = 0;
}
#else

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include "unzip.h"
#define DSP_DEVICE "/dev/dsp"

struct pcm pcm;

static int dsp;
static char *dsp_device;
static int stereo = 0;
static int samplerate = 44100;
static int sound = 1;

rcvar_t pcm_exports[] =
{
	RCV_BOOL("sound", &sound),
	RCV_INT("stereo", &stereo),
	RCV_INT("samplerate", &samplerate),
	RCV_STRING("oss_device", &dsp_device),
	RCV_END
};


void pcm_init()
{
	int n;

	if (!sound)
	{
		pcm.hz = 11025;
		pcm.len = 4096;
		pcm.buf = malloc(pcm.len);
		pcm.pos = 0;
		dsp = -1;
		return;
	}

	if (!dsp_device) dsp_device = strdup(DSP_DEVICE);
	dsp = open(dsp_device, O_WRONLY);

	n = 0x80009;
	ioctl(dsp, SNDCTL_DSP_SETFRAGMENT, &n);
/* 	n = AFMT_U8; */
	n = AFMT_S16_LE;
	ioctl(dsp, SNDCTL_DSP_SETFMT, &n);
	n = stereo;
	ioctl(dsp, SNDCTL_DSP_STEREO, &n);
	pcm.stereo = n;
	n = samplerate;
	ioctl(dsp, SNDCTL_DSP_SPEED, &n);
	pcm.hz = n;
	pcm.len = n / 60;
	pcm.buf = malloc(pcm.len * 2);
}

void pcm_close()
{
	if (pcm.buf) free(pcm.buf);
	memset(&pcm, 0, sizeof pcm);
	close(dsp);
}

void render_pcm(int n)
{
	if (dsp < 0)
	{
		pcm.pos = 0;
/* 		return 0; */
	}
	sound_mix(pcm.len - pcm.pos);
	if (pcm.buf) write(dsp, pcm.buf, pcm.pos * 2);
	pcm.pos = 0;
/* 	return 1; */
}

#endif /* USE_SDL_AUDIO */
